import base64
import copy
from .utils import dotdict
from uuid import UUID, uuid4
from bson import BSON
from pymongo.errors import OperationFailure


def _flatten_fetched_fields(fields_arg):
    """ this method takes either a kwargs 'fields', which can be a dict :
    {"_id": False, "store": 1, "url": 1} or a list : ["store", "flag", "url"]
    and returns a tuple : ("store", "flag").
    it HAS to be a tuple, so that it is not updated by the different instances
    """
    if fields_arg is None:
        return None
    if isinstance(fields_arg, dict):
        return tuple(sorted([k for k in fields_arg.keys() if fields_arg[k]]))
    else:
        return tuple(sorted(fields_arg))


class Document(dict):

    _initialized_with_doc = False
    _fetched_fields = None
    mongokat_collection = None
    gen_skel = True

    def __init__(self, doc=None, mongokat_collection=None, fetched_fields=None, gen_skel=None):

        if mongokat_collection is not None:
            self.mongokat_collection = mongokat_collection
        self.collection = self.mongokat_collection.collection

        if fetched_fields is not None:
            self._fetched_fields = fetched_fields
        self._fetched_fields = _flatten_fetched_fields(self._fetched_fields)

        if gen_skel is not None:
            self.gen_skel = gen_skel

        if doc is not None:
            for k, v in doc.iteritems():
                self[k] = v

        if not self._fetched_fields:
            self._initialized_with_doc = True

        if self.gen_skel:
            self.generate_skeleton()

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, dict(self))

    def __hash__(self):
        if '_id' in self:
            value = self['_id']
            return value.__hash__()
        else:
            raise TypeError("A Document is not hashable if it is not saved. Save the document before hashing it")

    def __deepcopy__(self, memo={}):
        obj = self.__class__(doc=copy.deepcopy(dict(self), memo), gen_skel=self.gen_skel, mongokat_collection=self.mongokat_collection, fetched_fields=self._fetched_fields)
        obj.__dict__ = self.__dict__.copy()
        return obj

    @property
    def b64id(self):
        """ Returns the document's _id as a base64-encoded string """
        return base64.b64encode(self["_id"].binary)

    def ensure_fields(self, fields, force_refetch=False):
        """ Makes sure we fetched the fields, and populate them if not. """

        # We fetched with fields=None, we should have fetched them all
        if self._fetched_fields is None or self._initialized_with_doc:
            return

        if force_refetch:
            missing_fields = fields
        else:
            missing_fields = [f for f in fields if f not in self._fetched_fields]

        if len(missing_fields) == 0:
            return

        if "_id" not in self:
            raise Exception("Can't ensure_fields because _id is missing")

        db_fields = self.mongokat_collection.find_one({"_id": self["_id"]}, fields={k: 1 for k in missing_fields})

        self._fetched_fields += tuple(missing_fields)

        if not db_fields:
            return

        for k, v in db_fields.iteritems():
            self[k] = v

    def unset_fields(self, fields):
        """ Removes this list of fields from both the local object and the DB. """

        for f in fields:
            if f in self:
                del self[f]

        self.mongokat_collection.update_one({"_id": self["_id"]}, {"$unset": {
            f: 1 for f in fields
        }})

    def reload(self):
        """
        allow to refresh the document, so after using update(), it could reload
        its value from the database.

        Be carreful : reload() will erase all unsaved values.

        If no _id is set in the document, a KeyError is raised.

        """

        old_doc = self.mongokat_collection.find_one({"_id": self['_id']}, read_use="primary")

        if not old_doc:
            raise OperationFailure('Can not reload an unsaved document.'
                                   ' %s is not found in the database. Maybe _id was a string and not ObjectId?' % self['_id'])
        else:
            for k in list(self.keys()):
                del self[k]
            self.update(dotdict(old_doc))

        self._initialized_with_doc = False

    def delete(self):
        """
        delete the document from the collection from his _id.
        """
        self.mongokat_collection.remove({'_id': self['_id']})

    def save(self, force=False, uuid=False, **kwargs):
        """
          REPLACES the object in DB. This is forbidden with objects from find() methods unless force=True is given.
        """

        if not self._initialized_with_doc and not force:
            raise Exception("Cannot save a document not initialized from a Python dict. This might remove fields from the DB!")

        self._initialized_with_doc = False

        if '_id' not in self:
            if uuid:
                self['_id'] = unicode("%s-%s" % (self.mongokat_collection.__class__.__name__, uuid4()))

        return self.mongokat_collection.save(self, **kwargs)

    def save_partial(self, data=None, allow_protected_fields=False, **kwargs):
        """ Saves just the currently set fields in the database. """

        # Backwards compat, deprecated argument
        if "dotnotation" in kwargs:
            del kwargs["dotnotation"]

        if data is None:

            data = dotdict(self)
            if "_id" not in data:
                raise KeyError("_id must be set in order to do a save_partial()")
            del data["_id"]

        if len(data) == 0:
          return

        if not allow_protected_fields:
            self.mongokat_collection._check_protected_fields(data)

        apply_on = dotdict(self)

        for k, v in data.iteritems():
            apply_on[k] = v

        self.update(dict(apply_on))

        self._initialized_with_doc = False

        self.mongokat_collection.update_one({"_id": self["_id"]}, {"$set": data}, **kwargs)

    def __generate_skeleton(self, doc, struct, path=""):

        for key in struct:
            if type(key) is type:
                new_key = "$%s" % key.__name__
            else:
                new_key = key
            new_path = ".".join([path, new_key]).strip('.')
            #
            # Automatique generate the skeleton with NoneType
            #
            if type(key) is not type and key not in doc:
                if isinstance(struct[key], dict):
                    if callable(struct[key]):
                        doc[key] = struct[key]()
                    else:
                        doc[key] = type(struct[key])()
                elif struct[key] is dict:
                    doc[key] = {}
                elif isinstance(struct[key], list):
                    doc[key] = type(struct[key])()
                # elif isinstance(struct[key], CustomType):
                #     if struct[key].init_type is not None:
                #         doc[key] = struct[key].init_type()
                #     else:
                #         doc[key] = None
                elif struct[key] is list:
                    doc[key] = []
                elif isinstance(struct[key], tuple):
                    doc[key] = [None for _ in range(len(struct[key]))]
                else:
                    doc[key] = None
            #
            # if the value is a dict, we have a another structure to validate
            #
            if isinstance(struct[key], dict) and type(key) is not type:
                self.__generate_skeleton(doc[key], struct[key], new_path)

    def generate_skeleton(self):
        if self.mongokat_collection.structure is not None:
            self.__generate_skeleton(self, self.mongokat_collection.structure)

    def get_size(self):
        """
        return the size of the underlying bson object
        """
        try:
            return len(BSON.encode(self))
        except:
            return None

    def validate(self):
        """ We do not support validation yet. """
        pass



from .utils import dotdict, json_clone
from bson import ObjectId
from bson.codec_options import CodecOptions
from pymongo import ReadPreference, WriteConcern, ReturnDocument, read_preferences
import collections
from .document import Document
from .exceptions import MultipleResultsFound, ImmutableDocumentError, ProtectedFieldsError


def _param_fields(kwargs, fields):
  """
    Normalize the "fields" argument to most find methods
  """
  if fields is None:
    return
  if type(fields) in [list, set, frozenset, tuple]:
    fields = {x: True for x in fields}
  if type(fields) == dict:
    fields.setdefault("_id", False)
  kwargs["projection"] = fields


def find_method(func):
  """
    Decorator that manages smart defaults or transforms for common find methods:

     - fields/projection: list of fields to be returned. Contrary to pymongo, _id won't be added automatically
     - json: performs a json_clone on the results. Beware of performance!
     - timeout
     - return_document
  """
  def wrapped(*args, **kwargs):

    # Normalize the fields argument if passed as a positional param.
    if len(args) == 3 and func.__name__ in ("find", "find_one", "find_by_id", "find_by_ids"):
      _param_fields(kwargs, args[2])
      args = (args[0], args[1])
    elif "fields" in kwargs:
      _param_fields(kwargs, kwargs["fields"])
      del kwargs["fields"]
    elif "projection" in kwargs:
      _param_fields(kwargs, kwargs["projection"])

    if "timeout" in kwargs:
      kwargs["no_cursor_timeout"] = not bool(kwargs["timeout"])
      del kwargs["timeout"]

    if "spec" in kwargs:
      kwargs["filter"] = kwargs["spec"]
      del kwargs["spec"]

    if kwargs.get("return_document") == "after":
        kwargs["return_document"] = ReturnDocument.AFTER
    elif kwargs.get("return_document") == "before":
        kwargs["return_document"] = ReturnDocument.BEFORE

    ret = func(*args, **kwargs)

    if kwargs.get("json"):
      ret = json_clone(ret)

    return ret

  return wrapped


def patch_cursor(cursor, batch_size=None, limit=None, skip=None, sort=None, **kwargs):
  """
    Adds batch_size, limit, sort parameters to a DB cursor
  """

  if type(batch_size) == int:
    cursor.batch_size(batch_size)

  if limit is not None:
    cursor.limit(limit)

  if sort is not None:
    cursor.sort(sort)

  if skip is not None:
    cursor.skip(skip)


class Collection(object):
    """ mongokat.Collection wraps a pymongo.collection.Collection """

    __collection__ = None
    __database__ = None
    document_class = Document
    structure = None
    immutable = False
    protected_fields = ()

    def __init__(self, collection=None, database=None, client=None):
        """ You can pass a pymongo collection object directly, or rely
            on the __collection__ and/or __database__ attributes
        """

        if collection:
            self.collection = collection
            self.database = collection.database
            self.client = self.database.client
        elif database and self.__collection__:
            self.database = database
            self.client = self.database.client
            self.collection = self.database[self.__collection__]
        elif client and self.__database__ and self.__collection__:
            self.client = client
            self.database = self.client[self.__database__]
            self.collection = self.database[self.__collection__]
        else:
            raise Exception("Not enough parameters given to identify the right collection!")

    def __call__(self, *args, **kwargs):
        """ Instanciates a new *Document* from this collection """
        kwargs["mongokat_collection"] = self
        return self.document_class(*args, **kwargs)

    #
    #
    # READ-ONLY METHODS
    #
    #

    def exists(self, query, **args):
        """
        Returns True if the search matches at least one document
        """
        return bool(self.find(query, **args).limit(1).count())

    def count(self, *args, **kwargs):
        return self._collection_with_options(kwargs).count(*args, **kwargs)

    def distinct(self, *args, **kwargs):
        return self._collection_with_options(kwargs).distinct(*args, **kwargs)

    def group(self, *args, **kwargs):
        return self._collection_with_options(kwargs).group(*args, **kwargs)

    @find_method
    def aggregate(self, *args, **kwargs):

        # Fix weird pymongo inconsistency https://github.com/mongodb/mongo-python-driver/blob/6865ba72edcda31c717037435e7985e9e4139dd9/test/test_crud.py#L85
        if "batch_size" in kwargs:
            kwargs["batchSize"] = kwargs["batch_size"]
            del kwargs["batch_size"]

        return self._collection_with_options(kwargs).aggregate(*args, **kwargs)

    @find_method
    def find(self, *args, **kwargs):
        return self._collection_with_options(kwargs).find(*args, **kwargs)

    def _collection_with_options(self, kwargs):
        """ Returns a copy of the pymongo collection with various options set up """

        # class DocumentClassWithFields(self.document_class):
        #     _fetched_fields = kwargs.get("projection")
        #     mongokat_collection = self

        read_preference = kwargs.get("read_preference") or getattr(self.collection, "read_preference", None) or ReadPreference.PRIMARY

        if "read_preference" in kwargs:
            del kwargs["read_preference"]

        # Simplified tag usage
        if "read_use" in kwargs:
            if kwargs["read_use"] == "primary":
                read_preference = ReadPreference.PRIMARY
            elif kwargs["read_use"] == "secondary":
                read_preference = ReadPreference.SECONDARY
            elif kwargs["read_use"] == "nearest":
                read_preference = ReadPreference.NEAREST
            elif kwargs["read_use"]:
                read_preference = read_preferences.Secondary(tag_sets=[{"use": kwargs["read_use"]}])
            del kwargs["read_use"]

        write_concern = None
        if kwargs.get("w") is 0:
            write_concern = WriteConcern(w=0)
        elif kwargs.get("write_concern"):
            write_concern = kwargs.get("write_concern")

        codec_options = CodecOptions(
            document_class=(
                self.document_class,
                {
                    "fetched_fields": kwargs.get("projection"),
                    "mongokat_collection": self
                }
            )
        )

        return self.collection.with_options(
            codec_options=codec_options,
            read_preference=read_preference,
            write_concern=write_concern
        )

    @find_method
    def find_one(self, *args, **kwargs):
        """
        Get a single document from the database.
        """
        doc = self._collection_with_options(kwargs).find_one(*args, **kwargs)
        if doc is None:
            return None

        return doc

    @find_method
    def find_by_id(self, _id, **kwargs):
        """
        Pass me anything that looks like an _id : str, ObjectId, {"_id": str}, {"_id": ObjectId}
        """

        if type(_id) == dict and _id.get("_id"):
            return self.find_one({"_id": ObjectId(_id["_id"])}, **kwargs)

        return self.find_one({"_id": ObjectId(_id)}, **kwargs)

    @find_method
    def find_by_ids(self, _ids, projection=None, **kwargs):
        """
            Does a big _id:$in query on any iterator
        """

        id_list = [ObjectId(_id) for _id in _ids]

        if len(_ids) == 0:
            return []  # FIXME : this should be an empty cursor !

        # Optimized path when only fetching the _id field.
        # Be mindful this might not filter missing documents that may not have been returned, had we done the query.
        if projection is not None and projection.keys() == ["_id"]:
            return [self({"_id": x}, fetched_fields={"_id": True}) for x in id_list]
        else:
            return self.find({"_id": {"$in": id_list}}, projection=projection, **kwargs)

    def list_column(self, *args, **kwargs):
        """
            Return one field as a list
        """
        return list(self.iter_column(*args, **kwargs))

    def iter_column(self, query=None, field="_id", **kwargs):
        """
            Return one field as an iterator.
            Beware that if your query returns records where the field is not set, it will raise a KeyError.
        """
        find_kwargs = {
            "projection": {"_id": False}
        }
        find_kwargs["projection"][field] = True

        cursor = self._collection_with_options(kwargs).find(query, **find_kwargs)  # We only want 1 field: bypass the ORM

        patch_cursor(cursor, **kwargs)

        return (dotdict(x)[field] for x in cursor)

    def find_random(self, **kwargs):
        """
        return one random document from the collection
        """
        import random
        max = self.count(**kwargs)
        if max:
            num = random.randint(0, max - 1)
            return self.find(**kwargs).skip(num).next()

    def one(self, *args, **kwargs):
        bson_obj = self.find(*args, **kwargs)
        count = bson_obj.count()
        if count > 1:
            raise MultipleResultsFound("%s results found" % count)
        elif count == 1:
            return bson_obj.next()

    #
    #
    # WRITE METHODS
    #
    #

    def insert(self, data, return_object=False):
        """ Inserts the data as a new document. """

        obj = self(data)  # pylint: disable=E1102
        obj.save()

        if return_object:
            return obj
        else:
            return obj["_id"]

    # http://api.mongodb.org/python/current/api/pymongo/collection.html

    def bulk_write(self, *args, **kwargs):
        """ Hook are not supported for this method! """
        return self.collection.bulk_write(*args, **kwargs)

    def insert_one(self, document, **kwargs):
        ret = self.collection.insert_one(document, **kwargs)
        self.trigger("after_save", ids=[ret.inserted_id], replacements=[document])
        return ret

    def insert_many(self, documents, **kwargs):
        ret = self.collection.insert_many(documents, **kwargs)
        self.trigger("after_save", ids=ret.inserted_ids, replacements=documents)
        return ret

    def replace_one(self, filter, replacement, **kwargs):

        if self.immutable:
            raise ImmutableDocumentError()

        if not kwargs.get("allow_protected_fields"):
            self._check_protected_fields(replacement)
        else:
            del kwargs["allow_protected_fields"]

        before_doc = None
        if self.has_trigger("after_save"):
            before_doc = self.find_one(filter, read_use="primary", projection=["_id"])

        ret = self.collection.replace_one(filter, replacement, **kwargs)

        if ret.modified_count is 0:
            return ret
        elif ret.upserted_id:
            self.trigger("after_save", replacements=[replacement], ids=[ret.upserted_id])
        elif before_doc:
            self.trigger("after_save", replacements=[replacement], ids=[before_doc["_id"]])

        return ret

    def update_one(self, filter, update, **kwargs):

        if self.immutable:
            raise ImmutableDocumentError()

        if "$set" in update:
            if not kwargs.get("allow_protected_fields"):
                self._check_protected_fields(update["$set"])
            else:
                del kwargs["allow_protected_fields"]

        before_doc = None
        if self.has_trigger("after_save"):
            before_doc = self.find_one(filter, read_use="primary", projection=["_id"])

        ret = self.collection.update_one(filter, update, **kwargs)

        if ret.modified_count is 0:
            return ret
        elif ret.upserted_id:
            self.trigger("after_save", update=update, ids=[ret.upserted_id])
        elif before_doc:
            self.trigger("after_save", update=update, ids=[before_doc["_id"]])

        return ret

    def update_many(self, filter, update, **kwargs):

        if self.immutable:
            raise ImmutableDocumentError()

        if "$set" in update:
            if not kwargs.get("allow_protected_fields"):
                self._check_protected_fields(update["$set"])
            else:
                del kwargs["allow_protected_fields"]

        before_ids = None
        if self.has_trigger("after_save"):
            before_ids = self.list_column(filter, read_use="primary")

        ret = self.collection.update_many(filter, update, **kwargs)

        if ret.modified_count is 0:
            return ret
        elif before_ids:
            self.trigger("after_save", ids=before_ids, update=update)

        return ret

    def delete_one(self, filter, **kwargs):
        doc = None
        if self.has_trigger("before_delete") or self.has_trigger("after_delete"):
            doc = self.find_one(filter, read_use="primary")
            self.trigger("before_delete", documents=[doc])

        ret = self.collection.delete_one(filter, **kwargs)

        if doc is not None:
            self.trigger("after_delete", documents=[doc])

        return ret

    def delete_many(self, filter, **kwargs):
        docs = []
        if self.has_trigger("before_delete") or self.has_trigger("after_delete"):
            docs = list(self.find(filter, read_use="primary"))
            self.trigger("before_delete", documents=docs)

        ret = self.collection.delete_many(filter, **kwargs)

        if len(docs) > 0:
            self.trigger("after_delete", documents=docs)

        return ret

    @find_method
    def find_one_and_delete(self, filter, **kwargs):
        self.trigger("before_delete", filter=filter)
        ret = self.collection.find_one_and_delete(filter, **kwargs)
        if ret is None:
            return None
        doc = self(ret, fetched_fields=kwargs.get("projection"))
        self.trigger("after_delete", documents=[doc])
        return doc

    @find_method
    def find_one_and_replace(self, filter, replacement, **kwargs):

        if self.immutable:
            raise ImmutableDocumentError()

        if not kwargs.get("allow_protected_fields"):
            self._check_protected_fields(replacement)
        else:
            del kwargs["allow_protected_fields"]

        ret = self.collection.find_one_and_replace(filter, replacement, **kwargs)
        if ret is None:
            return None
        doc = self(ret, fetched_fields=kwargs.get("projection"))
        self.trigger("after_save", documents=[doc], replacements=[replacement])
        return doc

    @find_method
    def find_one_and_update(self, filter, update, **kwargs):

        if self.immutable:
            raise ImmutableDocumentError()

        if "$set" in update:
            if not kwargs.get("allow_protected_fields"):
                self._check_protected_fields(update["$set"])
            else:
                del kwargs["allow_protected_fields"]

        ret = self.collection.find_one_and_update(filter, update, **kwargs)
        if ret is None:
            return None
        doc = self(ret, fetched_fields=kwargs.get("projection"))
        self.trigger("after_save", documents=[doc], update=update)
        return doc

    #
    #
    # EVENTS MANAGEMENT
    #
    #

    def has_trigger(self, event):
        """ Does this trigger need to run? """
        return hasattr(self.document_class, event)

    def trigger(self, event, filter=None, update=None, documents=None, ids=None, replacements=None):
        """ Trigger the after_save hook on documents, if present. """

        if not self.has_trigger(event):
            return

        if documents is not None:
            pass
        elif ids is not None:
            documents = self.find_by_ids(ids, read_use="primary")
        elif filter is not None:
            documents = self.find(filter, read_use="primary")
        else:
            raise Exception("Trigger couldn't filter documents")

        for doc in documents:
            getattr(doc, event)(update=update, replacements=replacements)
    #
    #
    # FOR BACKWARDS-COMPATIBILITY
    #
    #

    @property
    def connection(self):
        return self.client

    @property
    def db(self):
        return self.database

    def save(self, to_save, **kwargs):

        if self.immutable and "_id" in to_save:
            raise ImmutableDocumentError()

        if not kwargs.get("allow_protected_fields"):
            self._check_protected_fields(to_save)
        else:
            del kwargs["allow_protected_fields"]

        if "safe" in kwargs:
            kwargs["w"] = 0 if not kwargs["safe"] else 1
            del kwargs["safe"]
        _id = self.collection.save(to_save, **kwargs)

        self.trigger("after_save", replacements=[to_save], ids=[_id])
        return _id

    def update(self, spec, document, **kwargs):

        if self.immutable:
            raise ImmutableDocumentError()

        if "$set" in document:
            if not kwargs.get("allow_protected_fields"):
                self._check_protected_fields(document["$set"])
            else:
                del kwargs["allow_protected_fields"]

        before_ids = None
        if self.has_trigger("after_save"):
            before_ids = self.list_column(spec, read_use="primary")

        ret = self.collection.update(spec, document, **kwargs)
        self.trigger("after_save", ids=before_ids, update=document)
        return ret

    def remove(self, spec_or_id=None, **kwargs):
        docs = []

        if self.has_trigger("before_delete") or self.has_trigger("after_delete"):
            limit = 0
            if spec_or_id is None:
                filter = {}
            elif not isinstance(spec_or_id, collections.Mapping):
                filter = {"_id": spec_or_id}
            else:
                filter = spec_or_id
                limit = 1 if kwargs.get("multi") is False else 0

            docs = list(self.find(filter, read_use="primary", limit=limit))
            self.trigger("before_delete", documents=docs)

        ret = self.collection.remove(spec_or_id=spec_or_id, **kwargs)

        if len(docs) > 0:
            self.trigger("after_delete", documents=docs)

        return ret

    def find_and_modify(self, query={}, update=None, **kwargs):

        if self.immutable:
            raise ImmutableDocumentError()

        if "$set" in update:
            if not kwargs.get("allow_protected_fields"):
                self._check_protected_fields(update["$set"])
            else:
                del kwargs["allow_protected_fields"]

        ret = self.collection.find_and_modify(query=query, update=update, **kwargs)
        if ret is None:
            return None
        self.trigger("after_save", ids=[ret["_id"]], update=update)
        return self(ret, fetched_fields=kwargs.get("projection"))

    def get_from_id(self, _id):
        return self.find_one({"_id": _id})

    def fetch(self, spec=None, *args, **kwargs):
        """
        return all document which match the structure of the object
        `fetch()` takes the same arguments than the the pymongo.collection.find method.
        The query is launch against the db and collection of the object.
        """
        if spec is None:
            spec = {}
        for key in self.structure:
            if key in spec:
                if isinstance(spec[key], dict):
                    spec[key].update({'$exists': True})
            else:
                spec[key] = {'$exists': True}
        return self.find(spec, *args, **kwargs)

    def fetch_one(self, *args, **kwargs):
        """
        return one document which match the structure of the object
        `fetch_one()` takes the same arguments than the the pymongo.collection.find method.
        If multiple documents are found, raise a MultipleResultsFound exception.
        If no document is found, return None
        The query is launch against the db and collection of the object.
        """
        bson_obj = self.fetch(*args, **kwargs)
        count = bson_obj.count()
        if count > 1:
            raise MultipleResultsFound("%s results found" % count)
        elif count == 1:
            # return self(bson_obj.next(), fetched_fields=kwargs.get("projection"))
            return bson_obj.next()

    def _check_protected_fields(self, data):

        if len(self.protected_fields):
            forbidden_fields = set(data.keys()) & set(self.protected_fields)
            if len(forbidden_fields) > 0:
                raise ProtectedFieldsError("cannot set those keys without allow_protected_fields : %s" % forbidden_fields)




"""
Welcome to the API documentation for MongoKat.

Please see the [README on GitHub](https://github.com/pricingassistant/mongokat) for more info about MongoKat.
"""

import os

if os.getenv("MONGOKAT_DISABLE_CBSON"):
  from . import _bson
else:
  try:
    from . import _cbson as _bson
  except Exception:
    from . import _bson


import pymongo

# This is the only monkey-patch needed to use our own bson.decode_all function,
# which implements https://jira.mongodb.org/browse/PYTHON-175
pymongo.helpers.bson = _bson

import bson


# This other monkey-patch is needed to disable the type check on document_class, because
# we also can pass a tuple to use additional kwargs in the document_class instanciation.
class CodecOptionsWithoutCheck(bson.codec_options.CodecOptions):
  def __new__(cls, document_class=dict,
              tz_aware=False, uuid_representation=bson.codec_options.PYTHON_LEGACY):
      # if not issubclass(document_class, MutableMapping):
      #     raise TypeError("document_class must be dict, bson.son.SON, or "
      #                     "another subclass of collections.MutableMapping")
      if not isinstance(tz_aware, bool):
          raise TypeError("tz_aware must be True or False")
      if uuid_representation not in bson.codec_options.ALL_UUID_REPRESENTATIONS:
          raise ValueError("uuid_representation must be a value "
                           "from bson.binary.ALL_UUID_REPRESENTATIONS")

      return tuple.__new__(
          cls, (document_class, tz_aware, uuid_representation))


bson.codec_options.CodecOptions = CodecOptionsWithoutCheck

from .collection import Collection, find_method
from .document import Document



from setuptools import setup  # , find_packages
import os
import re
import sys
import platform
import warnings

from distutils.cmd import Command
from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsOptionError
from distutils.errors import DistutilsPlatformError, DistutilsExecError
from distutils.core import Extension


# PYTHON-654 - Clang doesn't support -mno-fused-madd but the pythons Apple
# ships are built with it. This is a problem starting with Xcode 5.1
# since clang 3.4 errors out when it encounters unrecognized compiler
# flags. This hack removes -mno-fused-madd from the CFLAGS automatically
# generated by distutils for Apple provided pythons, allowing C extension
# builds to complete without error. The inspiration comes from older
# versions of distutils.sysconfig.get_config_vars.
if sys.platform == 'darwin' and 'clang' in platform.python_compiler().lower():
    from distutils.sysconfig import get_config_vars
    res = get_config_vars()
    for key in ('CFLAGS', 'PY_CFLAGS'):
        if key in res:
            flags = res[key]
            flags = re.sub('-mno-fused-madd', '', flags)
            res[key] = flags

if sys.platform == 'win32' and sys.version_info > (2, 6):
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    build_errors = (CCompilerError, DistutilsExecError,
                    DistutilsPlatformError, IOError)
else:
    build_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)


class custom_build_ext(build_ext):
    """Allow C extension building to fail.

    The C extension speeds up BSON encoding, but is not essential.
    """

    warning_message = """
********************************************************************
WARNING: %s could not
be compiled. No C extensions are essential for PyMongo to run,
although they do result in significant speed improvements.
%s

Please see the installation docs for solutions to build issues:

http://api.mongodb.org/python/current/installation.html

Here are some hints for popular operating systems:

If you are seeing this message on Linux you probably need to
install GCC and/or the Python development package for your
version of Python.

Debian and Ubuntu users should issue the following command:

    $ sudo apt-get install build-essential python-dev

Users of Red Hat based distributions (RHEL, CentOS, Amazon Linux,
Oracle Linux, Fedora, etc.) should issue the following command:

    $ sudo yum install gcc python-devel

If you are seeing this message on Microsoft Windows please install
PyMongo using the MS Windows installer for your version of Python,
available on pypi here:

http://pypi.python.org/pypi/pymongo/#downloads

If you are seeing this message on OSX please read the documentation
here:

http://api.mongodb.org/python/current/installation.html#osx
********************************************************************
"""

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            e = sys.exc_info()[1]
            sys.stdout.write('%s\n' % str(e))
            warnings.warn(self.warning_message % ("Extension modules",
                                                  "There was an issue with "
                                                  "your platform configuration"
                                                  " - see above."))

    def build_extension(self, ext):
        name = ext.name
        if sys.version_info[:3] >= (2, 6, 0):
            try:
                build_ext.build_extension(self, ext)
            except build_errors:
                e = sys.exc_info()[1]
                sys.stdout.write('%s\n' % str(e))
                warnings.warn(self.warning_message % ("The %s extension "
                                                      "module" % (name,),
                                                      "The output above "
                                                      "this warning shows how "
                                                      "the compilation "
                                                      "failed."))
        else:
            warnings.warn(self.warning_message % ("The %s extension "
                                                  "module" % (name,),
                                                  "Please use Python >= 2.6 "
                                                  "to take advantage of the "
                                                  "extension."))

ext_modules = [Extension('mongokat._cbson',
                         include_dirs=['mongokat/_bson'],
                         sources=['mongokat/_bson/_cbsonmodule.c',
                                  'mongokat/_bson/time64.c',
                                  'mongokat/_bson/buffer.c',
                                  'mongokat/_bson/encoding_helpers.c'])
               ]

extra_opts = {}

if "--no_ext" in sys.argv:
    sys.argv.remove("--no_ext")
elif (sys.platform.startswith("java") or
      sys.platform == "cli" or
      "PyPy" in sys.version):
    sys.stdout.write("""
*****************************************************\n
The optional C extensions are currently not supported\n
by this python implementation.\n
*****************************************************\n
""")
elif sys.byteorder == "big":
    sys.stdout.write("""
*****************************************************\n
The optional C extensions are currently not supported\n
on big endian platforms and will not be built.\n
Performance may be degraded.\n
*****************************************************\n
""")
else:
    extra_opts['ext_modules'] = ext_modules


def get_requirements():
    reqs = []
    for filename in ["requirements.txt"]:
        with open(filename, "r") as f:
            reqs += [x.strip() for x in f.readlines() if x.strip() and not x.strip().startswith("#")]
    return reqs


def get_version():
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, 'mongokat/version.py')) as f:
        locals = {}
        exec(f.read(), locals)
        return locals['VERSION']
    raise RuntimeError('No version info found.')


try:
    long_desc = open("docs/index.rst", "").read()
except:
    long_desc = ""

setup(
    name="mongokat",
    include_package_data=True,
    packages=['mongokat'],
    version=get_version(),
    description="Lightweight MongoDB ORM/ODM",
    author="Pricing Assistant",
    license='MIT',
    author_email="contact@pricingassistant.com",
    url="http://github.com/pricingassistant/mongokat",
    # download_url="http://chardet.feedparser.org/download/python3-chardet-1.0.1.tgz",
    keywords=["mongodb", "mongo", "orm", "odm", "mongokit"],
    platforms='any',
    entry_points={
        'console_scripts': [
        ]
    },
    # dependency_links=[
    #     "http://github.com/mongodb/mongo-python-driver/archive/cb4adb2193a83413bc5545d89b7bbde4d6087761.zip#egg=pymongo-2.7rc1"
    # ],
    zip_safe=False,
    cmdclass={"build_ext": custom_build_ext},
    install_requires=get_requirements(),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        #'Development Status :: 1 - Planning',
        #'Development Status :: 2 - Pre-Alpha',
        #'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        #'Development Status :: 6 - Mature',
        #'Development Status :: 7 - Inactive',
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Utilities"
    ],
    long_description=long_desc,
    **extra_opts
)



import copy
import json
from bson import ObjectId
import datetime

try:
  from requests.structures import CaseInsensitiveDict
except:
  CaseInsensitiveDict = None


class _CustomJsonEncoder(json.JSONEncoder):
  def default(self, obj):  # pylint: disable-msg=E0202
    if isinstance(obj, datetime.datetime):
      return obj.isoformat()
    elif isinstance(obj, ObjectId):
      return str(obj)
    elif CaseInsensitiveDict and isinstance(obj, CaseInsensitiveDict):
      return dict(obj)
    elif type(obj) == set:
      return list(obj)
    return json.JSONEncoder.default(self, obj)

_CUSTOM_JSON_ENCODER = _CustomJsonEncoder()


def json_clone(obj):
  return json.loads(_CUSTOM_JSON_ENCODER.encode(obj))


class dotdict(dict):
  """"
  >>> life = dotdict({'bigBang': {'stars': {'planets': {}}}})
  >>> life.bigBang.stars.planets
  {}
  >>> life.bigBang.stars.planets.earth = { 'singleCellLife' : {} }
  >>> life.bigBang.stars.planets
  {'earth': {'singleCellLife': {}}}
  >>> life['bigBang.stars.planets.mars.landers.vikings'] = 2
  >>> life.bigBang.stars.planets.mars.landers.vikings
  2
  >>> 'landers.vikings' in life.bigBang.stars.planets.mars
  True
  >>> life.get('bigBang.stars.planets.mars.landers.spirit', True)
  True
  >>> life.setdefault('bigBang.stars.planets.mars.landers.opportunity', True)
  True
  >>> 'landers.opportunity' in life.bigBang.stars.planets.mars
  True
  >>> life.bigBang.stars.planets.mars
  {'landers': {'opportunity': True, 'vikings': 2}}
  """

  def __init__(self, value=None):
    if value is None:
      pass
    elif isinstance(value, dict):
      for key in value:
        self.__setitem__(key, value[key])
    else:
      raise TypeError('expected dict')

  def __setitem__(self, key, value):
    if '.' in key:
      myKey, restOfKey = key.split('.', 1)
      target = self.setdefault(myKey, dotdict())
      if not isinstance(target, dotdict):
        raise KeyError('cannot set "%s" in "%s" (%s)' % (restOfKey, myKey, repr(target)))
      target[restOfKey] = value
    else:
      if isinstance(value, dict) and not isinstance(value, dotdict):
        value = dotdict(value)
      dict.__setitem__(self, key, value)

  def __getitem__(self, key):
    if '.' not in key:
      return dict.__getitem__(self, key)
    myKey, restOfKey = key.split('.', 1)
    target = dict.__getitem__(self, myKey)
    if not isinstance(target, dotdict):
      raise KeyError('cannot get "%s" in "%s" (%s)' % (restOfKey, myKey, repr(target)))
    return target[restOfKey]

  def __contains__(self, key):
    if '.' not in key:
      return dict.__contains__(self, key)
    myKey, restOfKey = key.split('.', 1)
    if myKey in self:
      target = dict.__getitem__(self, myKey)
    else:
      return False
    if not isinstance(target, dotdict):
      return False
    return restOfKey in target

  def setdefault(self, key, default):
    if key not in self:
      self[key] = default
    return self[key]

  def __deepcopy__(self, memo):
    return copy.deepcopy(dict(self))
